// ----------------------------------------------------------------------------
// Project:	Implementation Alternatives for Bottom-Up Evaluation
// Filename:	test2pull.cpp
// Purpose:	Time Measurements for Rule Implementations: Pull method
// Last Change:	26.01.2010
// Language:	C++
// EMail:	brass@acm.org
// WWW:		http://www.informatik.uni-halle.de/~brass/
// Address:	Univ. Halle, Von-Seckendorff-Platz 1, D-06120 Halle, GERMANY
// Copyright:	(c) 2010 by Stefan Brass
//		Downloads and modifications for tests and scientific purposes
//		permitted, provided that some acknowledgement of my work
//		remains, and modified versions are clearly marked as such.
//		There is no warranty at all - this code may contain bugs.
// ----------------------------------------------------------------------------

// Example Program 2:
// p3(X, Y) :- q(X, Y), p2(Y, Z).
// p2(X, Y) :- q(X, Y), p1(Y, Z).
// p1(X, Y) :- q(X, Y), p0(Y, Z).
// p0(X, Y) :- r(X, Y).
//
// q := {(i,i+1)|i = 1,...,200}
// r := {(i,i)  |i = 1,...,200}



#include <iostream>
using namespace std;

#if defined(_MSC_VER) || defined(_WINDOWS_)
#define WINDOWS
#else
#define UNIX
#endif

// ----------------------------------------------------------------------------
// Sizes for test:
// ----------------------------------------------------------------------------

const int Q_TUPLES = 200;

const int R_TUPLES = 200;

const bool PRINT_EXTENSION = false;

// ----------------------------------------------------------------------------
// Timing Functions (UNIX Version):
// ----------------------------------------------------------------------------

#ifdef UNIX
#include <sys/time.h>
#include <time.h>

class timer {
public:
	void start()
	{
		gettimeofday(&Time_Start, 0);
		Clock_Start = clock();
	}

	void end()
	{
		gettimeofday(&Time_End, 0);
		Clock_End = clock();
	}
	
	long int realtime_msec()
	{
		long sec = Time_End.tv_sec - Time_Start.tv_sec;
		long usec = Time_End.tv_usec - Time_Start.tv_usec;
		if(usec < 0) {
			usec += 1000000;
			sec -= 1;
		}
		return sec * 1000 + (usec + 500) / 1000;
	}
	long int cputime_msec()
	{
		long ticks = Clock_End - Clock_Start;
		return ticks / (CLOCKS_PER_SEC / 1000);
	}

private:
	timeval Time_Start, Time_End;
	clock_t Clock_Start, Clock_End;
};
#endif

// ----------------------------------------------------------------------------
// Timing Functions (Windows Version):
// ----------------------------------------------------------------------------

#ifdef WINDOWS
#include <windows.h>
#include <time.h>

class timer {
public:
	void start()
	{
		GetSystemTimeAsFileTime(&(Start.ft));
		Clock_Start = clock();
	}

	void end()
	{
		GetSystemTimeAsFileTime(&(End.ft));
		Clock_End = clock();
	}
	
	long int realtime_msec()
	{
		return static_cast<long> ((End.ns100 - Start.ns100) / 10000);
	}
	long int cputime_msec()
	{
		long ticks = Clock_End - Clock_Start;
		return ticks / (CLOCKS_PER_SEC / 1000);
	}

private:
	union {
		long long ns100;
		FILETIME ft;
	} Start, End;
	clock_t Clock_Start, Clock_End;
};
#endif


//-----------------------------------------------------------------------------
// Predicate q: Tuples generated by computation
//-----------------------------------------------------------------------------

class q_cursor {
public:
	q_cursor() { }

	void open() { i = 0; }

	bool fetch()
	{
		if(i < Q_TUPLES) {
			i++;
			return true;
		}
		else
			return false;
	}

	int col_1() { return i; }

	int col_2() { return i+1; }

	void close() {}

private:
	int i;
};

//-----------------------------------------------------------------------------
// Predicate r: Tuples generated by computation
//-----------------------------------------------------------------------------

class r_cursor {
public:
	r_cursor() { }

	void open() { i = 0; }

	bool fetch()
	{
		if(i < R_TUPLES) {
			i++;
			return true;
		}
		else
			return false;
	}

	int col_1() { return i; }

	int col_2() { return i; }

	void close() {}

private:
	int i;
};

// ----------------------------------------------------------------------------
// Predicate p0 - Pull Method
// ----------------------------------------------------------------------------

// p0(X, Y) :- r(X, Y).

class p0_cursor {
public:
    void open()
    {
        r_c.open();
    }
    bool fetch()
    {
	    return r_c.fetch();
    }
    int col_1() { return r_c.col_1(); }
    int col_2() { return r_c.col_2(); }
    void close()
    {
            r_c.close();
    }
private:
    r_cursor r_c;
};

// ----------------------------------------------------------------------------
// Predicate p1 - Pull Method
// ----------------------------------------------------------------------------

// p1(X, Z) :- q(X, Y), p0(Y, Z).

class p1_cursor {
public:
    void open()
    {
        left.open();
        left_more = left.fetch();
        if(left_more)
            right.open();
    }
    bool fetch()
    {
        while(left_more) {
            while(right.fetch()) {
                if(left.col_2() == right.col_1())
                    return true;
            }
            right.close();
            left_more = left.fetch();
            if(left_more)
                right.open();
        }
        return false;
    }
    int col_1() { return left.col_1(); }
    int col_2() { return right.col_2(); }
    void close()
    {
        left.close();
        if(left_more)
            right.close();
    }
private:
    q_cursor left;
    p0_cursor right;
    bool left_more;
};


// ----------------------------------------------------------------------------
// Predicate p2 - Pull Method
// ----------------------------------------------------------------------------

// p2(X, Z) :- q(X, Y), p1(Y, Z).

class p2_cursor {
public:
    void open()
    {
        left.open();
        left_more = left.fetch();
        if(left_more)
            right.open();
    }
    bool fetch()
    {
        while(left_more) {
            while(right.fetch()) {
                if(left.col_2() == right.col_1())
                    return true;
            }
            right.close();
            left_more = left.fetch();
            if(left_more)
                right.open();
        }
        return false;
    }
    int col_1() { return left.col_1(); }
    int col_2() { return right.col_2(); }
    void close()
    {
        left.close();
        if(left_more)
            right.close();
    }
private:
    q_cursor left;
    p1_cursor right;
    bool left_more;
};


// ----------------------------------------------------------------------------
// Predicate p3 - Pull Method
// ----------------------------------------------------------------------------

// p3(X, Z) :- q(X, Y), p2(Y, Z).

class p3_cursor {
public:
    void open()
    {
        left.open();
        left_more = left.fetch();
        if(left_more)
            right.open();
    }
    bool fetch()
    {
        while(left_more) {
            while(right.fetch()) {
                if(left.col_2() == right.col_1())
                    return true;
            }
            right.close();
            left_more = left.fetch();
            if(left_more)
                right.open();
        }
        return false;
    }
    int col_1() { return left.col_1(); }
    int col_2() { return right.col_2(); }
    void close()
    {
        left.close();
        if(left_more)
            right.close();
    }
private:
    q_cursor left;
    p2_cursor right;
    bool left_more;
};


//-----------------------------------------------------------------------------
// Main Program:
//-----------------------------------------------------------------------------

int main()
{
	timer t;
	int c1 = 0;
	int c2 = 0;

	cout << "Test of Method 2: Pull [Program 2]\n";
	t.start();
	p3_cursor p3_c;
	p3_c.open();
	while(p3_c.fetch()) {
		c1 += p3_c.col_1();
		c2 += p3_c.col_2();
	}
	p3_c.close();
	t.end();

	if(PRINT_EXTENSION) {
		int num_tuples = 0;
		cout << "Extension of p3:\n\n";
		p3_c.open();
		while(p3_c.fetch()) {
			num_tuples++;
			cout << "p3(" << p3_c.col_1() << ", " << p3_c.col_2()
				<< ").\n";
		}
		cout << "\n\nTotal number of tuples: " << num_tuples << "\n";
		p3_c.close();
	}

	cout << "Column sums: col1=" << c1 << ", col2=" << c2 << "\n";
	cout << "\nUsed time: " << t.realtime_msec() << " ms,  "
		<< "CPU: " << t.cputime_msec() << " ms.\n\n\n";

	return 0;
}


